
# auto generated
class VersionInfo(object):
    @property
    def build_date(self):
      return "2025-06-10 08:16:54"

    @property
    def version(self):
      return "0.2.4"

    @property
    def build_user(self):
      return "ganrunsheng"
