# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import os
from aworld.metrics.context_manager import MetricContext

# MetricContext.configure(provider="otlp",
#                         backend="logfire",
#                         write_token=os.getenv("LOGFIRE_WRITE_TOKEN")
# )
